"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const limiter_1 = require("limiter");
const google_maps_services_js_1 = require("@googlemaps/google-maps-services-js");
const tyranid_1 = require("tyranid");
const appError_1 = require("../../core/appError");
const client = new google_maps_services_js_1.Client({});
const limiter = new limiter_1.RateLimiter(5, 'second');
const asyncRemoveTokens = (count) => {
    return new Promise((resolve, reject) => {
        limiter.removeTokens(count, (error, remainingRequests) => {
            if (error)
                return reject(error);
            resolve(remainingRequests);
        });
    });
};
const ensureApiKey = () => {
    var _a, _b;
    const apiKey = (_b = (_a = tyranid_1.Tyr.options.google) === null || _a === void 0 ? void 0 : _a.map) === null || _b === void 0 ? void 0 : _b.apiKey;
    if (!apiKey) {
        throw new appError_1.AppError('You need to set up your Google Map API key and provide it to Tyranid');
    }
    return apiKey;
};
exports.geocode = async (address, address2) => {
    const apiKey = ensureApiKey();
    try {
        await asyncRemoveTokens(1);
        let response = await client.geocode({
            params: {
                address,
                key: apiKey,
            },
        });
        if (response.data.status === 'OK') {
            return response.data.results[0].geometry.location;
        }
        console.log(`No geocode response for address: ${address}`, JSON.stringify(response, null, 2));
        if (address2) {
            await asyncRemoveTokens(1);
            response = await client.geocode({
                params: {
                    address: address2,
                    key: apiKey,
                },
            });
            if (response.data.status === 'OK') {
                return response.data.results[0].geometry.location;
            }
        }
        console.log(`No geocode response for address2: ${address2}`, JSON.stringify(response, null, 2));
        return null;
    }
    catch (err) {
        throw new appError_1.AppError(err);
    }
};
exports.reverseGeocode = async (latitude, longitude) => {
    const apiKey = ensureApiKey();
    try {
        await asyncRemoveTokens(1);
        const response = await client.reverseGeocode({
            params: {
                latlng: [latitude, longitude],
                key: apiKey,
            },
        });
        if (response.data.status === 'OK') {
            return response.data.results[0].formatted_address;
        }
        console.log(`No reverse geocode response for address: [${latitude},${longitude}]`, JSON.stringify(response, null, 2));
    }
    catch (err) {
        throw new appError_1.AppError(err);
    }
    return null;
};
exports.batchGeocode = async (addresses) => {
    const data = [];
    await Promise.all(addresses.map((address, idx) => {
        return exports.geocode(address)
            .then(latLong => {
            data[idx] = latLong;
        })
            .catch(err => {
            data[idx] = err;
        });
    }));
    return data;
};
exports.distance = async (origin, destination, miles) => {
    const apiKey = ensureApiKey();
    try {
        await asyncRemoveTokens(1);
        const response = await client.distancematrix({
            params: {
                origins: [origin],
                destinations: [destination],
                units: miles ? google_maps_services_js_1.UnitSystem.imperial : google_maps_services_js_1.UnitSystem.metric,
                key: apiKey,
            },
        });
        if (response.data.status === 'OK') {
            return response.data.rows[0].elements[0].distance.value;
        }
        console.log(`No distance for address: [${origin},${destination}]`, JSON.stringify(response, null, 2));
        const errMsg = response.data.error_message;
        if (errMsg) {
            console.log(response.data.status + ':' + errMsg);
        }
        else {
            console.log('Error: no response status!');
        }
    }
    catch (err) {
        throw new appError_1.AppError(err);
    }
};
tyranid_1.Tyr.google = {
    geocode: exports.geocode,
    reverseGeocode: exports.reverseGeocode,
    batchGeocode: exports.batchGeocode,
    distance: exports.distance,
};
//# sourceMappingURL=service.js.map